import sys
import tarfile
import tempfile
from pathlib import Path

import scipy.io

from .__about__ import __version__
from .main import show, write_png


def _get_version_text():
    return "\n".join(
        [
            "betterspy {} [Python {}.{}.{}]".format(
                __version__,
                sys.version_info.major,
                sys.version_info.minor,
                sys.version_info.micro,
            ),
            "Copyright (c) 2018-2020 Nico Schlömer <nico.schloemer@gmail.com>",
        ]
    )


def _read_matrix(filename):
    return {".mtx": scipy.io.mmread, ".mm": scipy.io.mmread, ".rb": scipy.io.hb_read}[
        filename.suffix
    ](filename)


def main(argv=None):
    # Parse command line arguments.
    parser = _get_parser()
    args = parser.parse_args(argv)

    infile = args.infile
    if infile.suffixes == [".tar", ".gz"]:
        with tarfile.open(infile, "r:gz") as tar:
            A = None
            for m in tar.getmembers():
                if Path(m.name).suffix in [".mtx", ".rb"]:
                    with tempfile.TemporaryDirectory() as tmpdir:
                        tar.extract(m, path=tmpdir)
                        filename = Path(tmpdir) / Path(m.name)
                        A = _read_matrix(filename)
                    break
            assert A is not None, f"Couldn't find matrix file in {infile}."
    else:
        A = _read_matrix(infile)

    if args.outfile is None:
        show(A)
    else:
        write_png(args.outfile, A, args.border_width, args.border_color, args.colormap)


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(
        description=("Show sparsity structure of matrix"),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("infile", type=Path, help="input matrix market file")

    parser.add_argument(
        "outfile", type=str, nargs="?", default=None, help="output png file (optional)"
    )

    parser.add_argument(
        "--border-width",
        "-w",
        required=False,
        type=int,
        default=0,
        help="border width (default: 0)",
    )

    parser.add_argument(
        "--border-color",
        "-b",
        required=False,
        type=str,
        default="0.5",
        help="border color (default: 0.5, gray)",
    )

    parser.add_argument(
        "--colormap",
        "-c",
        required=False,
        type=str,
        default=None,
        help="border color (default: 0.5, gray)",
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=_get_version_text(),
        help="display version information",
    )
    return parser



import tempfile
from pathlib import Path

import imageio
import numpy
import pytest
from scipy import sparse

import betterspy


def test_show():
    M = sparse.rand(20, 20, density=0.1)
    betterspy.show(M)


@pytest.mark.parametrize(
    "ref, kwargs",
    [
        (6875310, {}),
        (7524085, {"border_width": 1}),
        (21306270, {"border_width": 1, "border_color": "red"}),
        (4981037, {"colormap": "viridis"}),
        (7101351, {"colormap": "viridis", "border_width": 1}),
    ],
)
def test_png(ref, kwargs):
    M = sparse.rand(20, 30, density=0.1, random_state=123)
    numpy.random.seed(123)

    with tempfile.TemporaryDirectory() as temp_dir:
        filepath = Path(temp_dir) / "test.png"
        betterspy.write_png(filepath, M, **kwargs)
        im = imageio.imread(filepath)
        y = numpy.random.randint(0, 100, size=numpy.prod(im.shape))
        assert numpy.dot(y, im.flatten()) == ref


def test_readme_images():
    import meshzoo
    from skfem import BilinearForm, ElementTriP2, InteriorBasis, MeshTri
    from skfem.helpers import dot, grad

    @BilinearForm
    def laplace(u, v, _):
        return dot(grad(u), grad(v))

    points, cells = meshzoo.rectangle_tri((-1.0, -1.0), (1.0, 1.0), 20)
    mesh = MeshTri(points.T, cells.T)

    basis = InteriorBasis(mesh, ElementTriP2())
    A = laplace.assemble(basis)

    with tempfile.TemporaryDirectory() as temp_dir:
        filepath = Path(temp_dir) / "test.png"
        betterspy.write_png(filepath, A, border_width=2)

    # betterspy.write_png(
    #     'ATA.png', M, border_width=2,
    #     colormap='viridis'
    #     )


def test_cli():
    this_dir = Path(__file__).resolve().parent
    mmfile = this_dir / "data" / "gre_343_343_crg.mm"
    betterspy.cli.main([mmfile.as_posix()])
    betterspy.cli.main([mmfile.as_posix(), "out.png"])


if __name__ == "__main__":
    test_readme_images()



"""
Better spy() function for Scipy sparse matrices.
"""
import tempfile

import matplotlib.colors as colors
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy
import png  # purepng


def plot(A, border_width=0, border_color="0.5", colormap=None):
    with tempfile.NamedTemporaryFile() as fp:
        write_png(
            fp.name,
            A,
            border_width=border_width,
            border_color=border_color,
            colormap=colormap,
        )
        img = mpimg.imread(fp.name)
        plt.imshow(img, origin="upper", interpolation="nearest")


def show(*args, **kwargs):
    plot(*args, **kwargs)
    plt.show()


class RowIterator:
    def __init__(self, A, border_width, border_color, colormap):
        self.A = A.tocsr()
        self.border_width = border_width

        rgb = numpy.array(colors.to_rgb(border_color))
        border_color_is_bw = numpy.all(rgb[0] == rgb) and rgb[0] in [0, 1]
        border_color_is_gray = numpy.all(rgb[0] == rgb)

        if colormap is None and (border_width == 0 or border_color_is_bw):
            self.mode = "binary"
            self.border_color = False
            self.bitdepth = 1
            self.dtype = bool

        elif colormap is None and border_color_is_gray:
            self.mode = "grayscale"
            self.bitdepth = 8
            self.dtype = numpy.uint8
            self.border_color = numpy.uint8(numpy.round(rgb[0] * 255))

        else:
            self.mode = "rgb"
            self.border_color = numpy.round(rgb * 255).astype(numpy.uint8)
            self.dtype = numpy.uint8
            self.bitdepth = 8

        if colormap is None:
            if self.mode == "binary":

                def convert_values(idx, vals):
                    out = numpy.ones(self.A.shape[1], dtype=self.dtype)
                    out[idx] = False
                    return out

            elif self.mode == "grayscale":

                def convert_values(idx, vals):
                    out = numpy.full(self.A.shape[1], 255, dtype=self.dtype)
                    out[idx] = 0
                    return out

            else:
                assert self.mode == "rgb"

                def convert_values(idx, vals):
                    out = numpy.full((self.A.shape[1], 3), 255, dtype=self.dtype)
                    out[idx, :] = 0
                    return out.flatten()

        else:
            assert self.mode == "rgb"
            # Convert the string into a colormap object with `to_rgba()`,
            # <https://stackoverflow.com/a/15140118/353337>.
            import matplotlib.cm as cmx

            cm = plt.get_cmap(colormap)
            c_norm = colors.Normalize(
                vmin=min(0.0, self.A.data.min()), vmax=max(0.0, self.A.data.max())
            )
            scalar_map = cmx.ScalarMappable(norm=c_norm, cmap=cm)

            def convert_values(idx, vals):
                x = numpy.zeros(self.A.shape[1])
                x[idx] = vals
                out = scalar_map.to_rgba(x)[:, :3] * 255
                out = numpy.round(out).astype(self.dtype)
                return out.flatten()

        self.convert_values = convert_values
        self.current = 0
        return

    def __iter__(self):
        return self

    def __next__(self):
        m = self.A.shape[0]
        b = self.border_width

        if self.current >= m + 2 * b:
            raise StopIteration

        if b == 0:
            row = self.A[self.current]
            out = self.convert_values(row.indices, row.data)
        else:
            if self.current < b or self.current > m + b - 1:
                out = numpy.tile(self.border_color, self.A.shape[1] + 2 * b).astype(
                    self.dtype
                )
            else:
                row = self.A[self.current - b]
                border = numpy.tile(self.border_color, b)
                out = numpy.concatenate(
                    [border, self.convert_values(row.indices, row.data), border]
                )

        self.current += 1
        return out


def write_png(filename, A, border_width=0, border_color="0.5", colormap=None):
    iterator = RowIterator(A, border_width, border_color, colormap)

    m, n = A.shape
    w = png.Writer(
        n + 2 * border_width,
        m + 2 * border_width,
        greyscale=iterator.mode != "rgb",
        bitdepth=iterator.bitdepth,
    )

    with open(filename, "wb") as f:
        w.write(f, iterator)
